@@ -105,6 +105,34 @@ type_from_str (const char *type)
 	return WSTYLUS_UNKNOWN;
 }
 
+static const char *
+str_from_type (WacomStylusType type)
+{
+	switch (type) {
+	case WSTYLUS_UNKNOWN:
+		return NULL;
+	case WSTYLUS_GENERAL:
+		return "General";
+	case WSTYLUS_INKING:
+		return "Inking";
+	case WSTYLUS_AIRBRUSH:
+		return "Airbrush";
+	case WSTYLUS_CLASSIC:
+		return "Classic";
+	case WSTYLUS_MARKER:
+		return "Marker";
+	case WSTYLUS_STROKE:
+		return "Stroke";
+	case WSTYLUS_PUCK:
+		return "Pick";
+	case WSTYLUS_3D:
+		return "3D";
+	case WSTYLUS_MOBILE:
+		return "Mobile";
+	}
+	return NULL;
+}
+
 static WacomEraserType
 eraser_type_from_str (const char *type)
 {
@@ -119,6 +147,22 @@ eraser_type_from_str (const char *type)
 	return WACOM_ERASER_UNKNOWN;
 }
 
+static const char *
+eraser_str_from_type (WacomEraserType type)
+{
+	switch (type) {
+	case WACOM_ERASER_NONE:
+		return "None";
+	case WACOM_ERASER_INVERT:
+		return "Invert";
+	case WACOM_ERASER_BUTTON:
+		return "Button";
+	case WACOM_ERASER_UNKNOWN:
+		return NULL;
+	}
+	return NULL;
+}
+
 WacomBusType
 bus_from_str (const char *str)
 {
@@ -253,8 +297,116 @@ libwacom_matchstr_to_paired(WacomDevice *device, const char *matchstr)
 	return TRUE;
 }
 
+static bool
+parse_stylus_id(const char *str, WacomStylusId *id)
+{
+	char **tokens = g_strsplit(str, ":", 2);
+	const char *vidstr, *tidstr;
+	int vid, tool_id;
+	bool ret = false;
+
+	if (tokens[1] == NULL) {
+		vidstr = G_STRINGIFY(WACOM_VENDOR_ID);
+		tidstr = tokens[0];
+	} else {
+		vidstr = tokens[0];
+		tidstr = tokens[1];
+	}
+
+	if (safe_atoi_base (vidstr, &vid, 16) && safe_atoi_base (tidstr, &tool_id, 16)) {
+		id->vid = vid;
+		id->tool_id = tool_id;
+		ret = true;
+	}
+	g_clear_pointer(&tokens, g_strfreev);
+
+	return ret;
+}
+
+static char *
+string_or_fallback (GKeyFile *keyfile,
+		    const char *group,
+		    const char *key,
+		    const char *fallback,
+		    GError *error)
+{
+	GError *local_error = NULL;
+	bool ret;
+
+	ret = g_key_file_has_key(keyfile, group, key, &local_error);
+	if (local_error) {
+		g_warning("String fallback error: %s", local_error->message);
+		g_clear_error(&local_error);
+	}
+	if (ret)
+		return g_key_file_get_string(keyfile, group, key, &error);
+
+	return g_strdup(fallback);
+}
+
+static int
+int_or_fallback (GKeyFile *keyfile,
+		 const char *group,
+		 const char *key,
+		 int fallback,
+		 GError *error)
+{
+	GError *local_error = NULL;
+	bool ret;
+
+	ret = g_key_file_has_key(keyfile, group, key, &local_error);
+	if (local_error) {
+		g_warning("int fallback error: %s", local_error->message);
+		g_clear_error(&local_error);
+	}
+
+	if (ret)
+		return g_key_file_get_integer(keyfile, group, key, &error);
+
+	return fallback;
+}
+
+static bool
+boolean_or_fallback (GKeyFile *keyfile,
+		     const char *group,
+		     const char *key,
+		     bool fallback,
+		     GError *error)
+{
+	GError *local_error = NULL;
+	bool ret;
+
+	ret = g_key_file_has_key(keyfile, group, key, &local_error);
+	if (local_error) {
+		g_warning("boolean fallback error: %s", local_error->message);
+		g_clear_error(&local_error);
+	}
+
+	if (ret)
+		return g_key_file_get_boolean(keyfile, group, key, &error);
+
+	return fallback;
+}
+
+static gchar **
+stylus_ids_as_hex(GArray *array)
+{
+	WacomStylusId *id;
+
+	if (array == NULL || array->len == 0)
+		return NULL;
+
+	GArray *a = g_array_new(true, false, array->len);
+	for (guint i = 0; i < array->len; i++) {
+		id = &g_array_index(array, WacomStylusId, i);
+		g_array_append_vals(a, g_strdup_printf("0x%08x", id->tool_id), 1);
+	}
+
+	return g_array_steal(a, NULL);
+}
+
 static void
-libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
+libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path, AliasStatus handle_aliases)
 {
 	GKeyFile *keyfile;
 	GError *error = NULL;
@@ -266,90 +418,141 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 
 	rc = g_key_file_load_from_file(keyfile, path, G_KEY_FILE_NONE, &error);
 	g_assert (rc);
+
 	groups = g_key_file_get_groups (keyfile, NULL);
 	for (i = 0; groups[i]; i++) {
-		WacomStylus *stylus;
+		WacomStylus *stylus = NULL, *aliased = NULL;
+		WacomStylusId id;
 		GError *error = NULL;
 		char *eraser_type, *type;
-		int id;
 		char **string_list;
 
-		if (!safe_atoi_base (groups[i], &id, 16)) {
-			g_warning ("Failed to parse stylus ID '%s'", groups[i]);
+		if (!parse_stylus_id(groups[i], &id)) {
+			g_warning ("Failed to parse stylus ID '%s', ignoring entry", groups[i]);
 			continue;
 		}
 
+		char *aliasstr = g_key_file_get_string(keyfile, groups[i], "AliasOf", NULL);
+
+		if (handle_aliases == IGNORE_ALIASES && aliasstr) {
+			g_clear_pointer(&aliasstr, g_free);
+			continue;
+		}
+
+		if (handle_aliases == ONLY_ALIASES && !aliasstr) {
+			g_clear_pointer(&aliasstr, g_free);
+			continue;
+		}
+
+		if (handle_aliases == ONLY_ALIASES) {
+			WacomStylusId alias_of = id;
+
+			/* Note: this effectively requires that all non-wacom AliasOf are specified
+			 * in the vid:pid format, otherwise they fall back to Wacom's vid. */
+			if (parse_stylus_id(aliasstr, &alias_of)) {
+				aliased = g_hash_table_lookup(db->stylus_ht, &alias_of);
+				if (!aliased) {
+					g_set_error(&error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
+						    "Unkown AliasOf %s reference, ignoring this entry", aliasstr);
+				}
+			} else {
+				g_set_error(&error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
+					    "Invalid AliasOf '%s', ignoring this entry", aliasstr);
+			}
+		}
+		if (error) {
+			g_warning ("[%s] %s", groups[i], error->message);
+			g_clear_error (&error);
+			g_clear_pointer(&aliasstr, g_free);
+			continue;
+		}
+		g_clear_pointer(&aliasstr, g_free);
+
 		stylus = g_new0 (WacomStylus, 1);
 		stylus->refcnt = 1;
 		stylus->id = id;
-		stylus->name = g_key_file_get_string(keyfile, groups[i], "Name", NULL);
-		stylus->group = g_key_file_get_string(keyfile, groups[i], "Group", NULL);
+		stylus->name = string_or_fallback(keyfile, groups[i], "Name", aliased ? aliased->name : NULL, NULL);
+		stylus->group = string_or_fallback(keyfile, groups[i], "Group", aliased ? aliased->group : NULL, NULL);
+		stylus->paired_stylus_ids = g_array_new (FALSE, FALSE, sizeof(WacomStylusId));
 
-		eraser_type = g_key_file_get_string(keyfile, groups[i], "EraserType", NULL);
+		eraser_type = string_or_fallback(keyfile, groups[i], "EraserType", aliased ? eraser_str_from_type(aliased->eraser_type) : NULL, NULL);
 		stylus->eraser_type = eraser_type_from_str (eraser_type);
 		g_clear_pointer(&eraser_type, g_free);
 
-		stylus->paired_ids = g_array_new (FALSE, FALSE, sizeof(int));
-		string_list = g_key_file_get_string_list (keyfile, groups[i], "PairedStylusIds", NULL, NULL);
-		for (guint j = 0; string_list && string_list[j]; j++) {
-			int val;
+		/* We have to keep the integer array for libwacom_get_supported_styli() */
+		stylus->deprecated_paired_ids = g_array_new (FALSE, FALSE, sizeof(int));
+		stylus->paired_styli = g_array_new (FALSE, FALSE, sizeof(WacomStylus*));
 
-			if (safe_atoi_base (string_list[j], &val, 16)) {
-				g_array_append_val (stylus->paired_ids, val);
+		string_list = g_key_file_get_string_list(keyfile, groups[i], "PairedStylusIds", NULL, NULL);
+		if (handle_aliases != IGNORE_ALIASES) {
+			if (string_list == NULL) {
+				string_list = stylus_ids_as_hex(aliased ? aliased->paired_stylus_ids: NULL);
+			}
+		}
+
+		for (guint j = 0; string_list && string_list[j]; j++) {
+			WacomStylusId paired_id;
+			if (parse_stylus_id(string_list[j], &paired_id)) {
+				g_array_append_val (stylus->paired_stylus_ids, paired_id);
+				if (paired_id.vid == WACOM_VENDOR_ID)
+					g_array_append_val (stylus->deprecated_paired_ids, paired_id.tool_id);
 			} else {
 				g_warning ("Stylus %s (%s) Ignoring invalid PairedStylusIds value\n", stylus->name, groups[i]);
 			}
 		}
 		g_clear_pointer(&string_list, g_strfreev);
 
-		stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", &error);
+		stylus->has_lens = boolean_or_fallback(keyfile, groups[i], "HasLens", aliased ? aliased->has_lens : FALSE, error);
 		if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
 			g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
 		g_clear_error (&error);
-		stylus->has_wheel = g_key_file_get_boolean(keyfile, groups[i], "HasWheel", &error);
+		stylus->has_wheel = boolean_or_fallback(keyfile, groups[i], "HasWheel", aliased ? aliased->has_wheel : FALSE, error);
 		if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
 			g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
 		g_clear_error (&error);
-
-		stylus->num_buttons = g_key_file_get_integer(keyfile, groups[i], "Buttons", &error);
+		stylus->num_buttons = int_or_fallback(keyfile, groups[i], "Buttons", aliased ? aliased->num_buttons : 0, error);
 		if (stylus->num_buttons == 0 && error != NULL) {
 			stylus->num_buttons = -1;
 			g_clear_error (&error);
 		}
 
-		stylus->axes = WACOM_AXIS_TYPE_NONE;
 		string_list = g_key_file_get_string_list (keyfile, groups[i], "Axes", NULL, NULL);
-		for (guint j = 0; string_list && string_list[j]; j++) {
-			WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
-			if (g_str_equal(string_list[j], "Tilt")) {
-				flag = WACOM_AXIS_TYPE_TILT;
-			} else if (g_str_equal(string_list[j], "RotationZ")) {
-				flag = WACOM_AXIS_TYPE_ROTATION_Z;
-			} else if (g_str_equal(string_list[j], "Distance")) {
-				flag = WACOM_AXIS_TYPE_DISTANCE;
-			} else if (g_str_equal(string_list[j], "Pressure")) {
-				flag = WACOM_AXIS_TYPE_PRESSURE;
-			} else if (g_str_equal(string_list[j], "Slider")) {
-				flag = WACOM_AXIS_TYPE_SLIDER;
-			} else {
-				g_warning ("Invalid axis %s for stylus ID %s\n",
-					   string_list[j], groups[i]);
+		if (handle_aliases == ONLY_ALIASES && (string_list == NULL)) {
+			stylus->axes = aliased->axes;
+		} else {
+			stylus->axes = WACOM_AXIS_TYPE_NONE;
+			for (guint j = 0; string_list && string_list[j]; j++) {
+				WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
+				if (g_str_equal(string_list[j], "Tilt")) {
+					flag = WACOM_AXIS_TYPE_TILT;
+				} else if (g_str_equal(string_list[j], "RotationZ")) {
+					flag = WACOM_AXIS_TYPE_ROTATION_Z;
+				} else if (g_str_equal(string_list[j], "Distance")) {
+					flag = WACOM_AXIS_TYPE_DISTANCE;
+				} else if (g_str_equal(string_list[j], "Pressure")) {
+					flag = WACOM_AXIS_TYPE_PRESSURE;
+				} else if (g_str_equal(string_list[j], "Slider")) {
+					flag = WACOM_AXIS_TYPE_SLIDER;
+				} else {
+					g_warning ("Invalid axis %s for stylus ID %s\n",
+						   string_list[j], groups[i]);
+			}
+				if (stylus->axes & flag)
+					g_warning ("Duplicate axis %s for stylus ID %s\n",
+						   string_list[j], groups[i]);
+				stylus->axes |= flag;
 			}
-			if (stylus->axes & flag)
-				g_warning ("Duplicate axis %s for stylus ID %s\n",
-					   string_list[j], groups[i]);
-			stylus->axes |= flag;
 		}
 		g_clear_pointer(&string_list, g_strfreev);
 
-		type = g_key_file_get_string(keyfile, groups[i], "Type", NULL);
+		type = string_or_fallback(keyfile, groups[i], "Type", aliased ? str_from_type(aliased->type) : NULL, NULL);
 		stylus->type = type_from_str (type);
 		g_clear_pointer(&type, g_free);
 
-		if (g_hash_table_lookup (db->stylus_ht, GINT_TO_POINTER (id)) != NULL)
-			g_warning ("Duplicate definition for stylus ID '%#x'", id);
+		if (g_hash_table_lookup (db->stylus_ht, &id) != NULL)
+			g_warning ("Duplicate definition for stylus ID '%s'", groups[i]);
 
-		g_hash_table_insert (db->stylus_ht, GINT_TO_POINTER (id), stylus);
+		g_hash_table_insert (db->stylus_ht, g_memdup2(&id, sizeof(id)), stylus);
 	}
 	g_clear_pointer(&groups, g_strfreev);
 	g_clear_pointer(&keyfile, g_key_file_free);
@@ -364,23 +567,24 @@ libwacom_setup_paired_attributes(WacomDeviceDatabase *db)
 	g_hash_table_iter_init(&iter, db->stylus_ht);
 	while (g_hash_table_iter_next (&iter, &key, &value)) {
 		WacomStylus *stylus = value;
-		const int* ids;
-		int count;
-		int i;
+		GArray *paired_ids = g_steal_pointer(&stylus->paired_stylus_ids);
 
-		ids = libwacom_stylus_get_paired_ids(stylus, &count);
-		for (i = 0; i < count; i++) {
-			const WacomStylus *pair;
+		for (guint i = 0; i < paired_ids->len; i++) {
+			WacomStylusId *id = &g_array_index(paired_ids, WacomStylusId, i);
+			WacomStylus *paired = g_hash_table_lookup(db->stylus_ht, id);
 
-			pair = libwacom_stylus_get_for_id(db, ids[i]);
-			if (pair == NULL) {
-				g_warning("Paired stylus '0x%x' not found, ignoring.", ids[i]);
+			if (paired == NULL) {
+				g_warning ("Ignoring paired stylus %04x:%x", id->vid, id->tool_id);
 				continue;
 			}
-			if (libwacom_stylus_is_eraser(pair)) {
+
+			g_array_append_val(stylus->paired_styli, paired);
+
+			if (libwacom_stylus_is_eraser(paired)) {
 				stylus->has_eraser = true;
 			}
 		}
+		g_array_unref(paired_ids);
 	}
 }
 
@@ -394,8 +598,8 @@ static const struct {
 	{ "Bottom", WACOM_BUTTON_POSITION_BOTTOM },
 	{ "Ring", WACOM_BUTTON_RING_MODESWITCH },
 	{ "Ring2", WACOM_BUTTON_RING2_MODESWITCH },
-	{ "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH },
-	{ "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH },
+	{ "Strip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH },
+	{ "Strip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH },
 	{ "OLEDs", WACOM_BUTTON_OLED },
 	{ "Dial", WACOM_BUTTON_DIAL_MODESWITCH },
 	{ "Dial2", WACOM_BUTTON_DIAL2_MODESWITCH },
@@ -407,8 +611,10 @@ static const struct {
 } supported_leds[] = {
 	{ "Ring",		WACOM_STATUS_LED_RING },
 	{ "Ring2",		WACOM_STATUS_LED_RING2 },
-	{ "Touchstrip",		WACOM_STATUS_LED_TOUCHSTRIP },
-	{ "Touchstrip2",	WACOM_STATUS_LED_TOUCHSTRIP2 }
+	{ "Strip",		WACOM_STATUS_LED_TOUCHSTRIP },
+	{ "Strip2",		WACOM_STATUS_LED_TOUCHSTRIP2 },
+	{ "Dial",		WACOM_STATUS_LED_DIAL },
+	{ "Dial2",		WACOM_STATUS_LED_DIAL2 },
 };
 
 static const struct {
@@ -416,7 +622,8 @@ static const struct {
 	WacomIntegrationFlags   value;
 } integration_flags[] = {
 	{ "Display",		WACOM_DEVICE_INTEGRATED_DISPLAY },
-	{ "System",		WACOM_DEVICE_INTEGRATED_SYSTEM }
+	{ "System",		WACOM_DEVICE_INTEGRATED_SYSTEM },
+	{ "Remote",		WACOM_DEVICE_INTEGRATED_REMOTE }
 };
 
 static void
@@ -427,10 +634,19 @@ libwacom_parse_buttons_key(WacomDevice      *device,
 {
 	guint i;
 	char **vals;
+	WacomModeSwitch mode = WACOM_MODE_SWITCH_NEXT;
 
 	vals = g_key_file_get_string_list (keyfile, BUTTONS_GROUP, key, NULL, NULL);
 	if (vals == NULL)
 		return;
+
+	/* If we have more than one entry our buttons
+	 * switch mode to a direct mode. Otherwise the
+	 * single button just cycles to next mode;
+	 */
+	if (vals[0] != NULL && vals[1] != NULL)
+		mode = WACOM_MODE_SWITCH_0;
+
 	for (i = 0; vals[i] != NULL; i++) {
 		char val;
 		WacomButton *button;
@@ -446,10 +662,20 @@ libwacom_parse_buttons_key(WacomDevice      *device,
 		button = g_hash_table_lookup(device->buttons, GINT_TO_POINTER(val));
 		if (!button) {
 			button = g_new0(WacomButton, 1);
+			button->mode = WACOM_MODE_SWITCH_NEXT;
 			g_hash_table_insert(device->buttons, GINT_TO_POINTER(val), button);
 		}
 
 		button->flags |= flag;
+
+		/* This is Good Enough. Devices with direct mode switch buttons
+		 * have those tied to a single feature only, e.g. Ring (and optionally Ring2).
+		 * We effectively take the order of the buttons in the .tablet file and
+		 * bind the buttons in that order to the mode - first button is mode 0, etc.
+		 * If there's only one button it is set to "next".
+		 */
+		if (flag & WACOM_BUTTON_MODESWITCH)
+			button->mode = mode++;
 	}
 
 	g_strfreev (vals);
@@ -688,12 +914,21 @@ libwacom_parse_keys(WacomDevice *device,
 	libwacom_parse_key_codes(device, keyfile);
 }
 
+static int
+wacom_stylus_id_sort(const WacomStylusId *a, const WacomStylusId *b)
+{
+	if (a->vid == b->vid)
+		return a->tool_id - b->tool_id;
+
+	return a->vid - b->vid;
+}
 
 static int
 styli_id_sort(gconstpointer pa, gconstpointer pb)
 {
-	const int *a = pa, *b = pb;
-	return *a > *b ? 1 : *a == *b ? 0 : -1;
+	const WacomStylus *a = *(WacomStylus**)pa, *b = *(WacomStylus**)pb;
+
+	return wacom_stylus_id_sort(&a->id, &b->id);
 }
 
 static void
@@ -703,17 +938,23 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 	GArray *array;
 	guint i;
 
-	array = g_array_new (FALSE, FALSE, sizeof(int));
-	for (i = 0; ids[i]; i++) {
-		const char *id = ids[i];
+	array = g_array_new (FALSE, FALSE, sizeof(WacomStylus*));
+	for (i = 0; ids && ids[i]; i++) {
+		const char *str = ids[i];
 
-		if (g_str_has_prefix(id, "0x")) {
-			int int_value;
-			if (safe_atoi_base (ids[i], &int_value, 16)) {
-				g_array_append_val (array, int_value);
+		if (g_str_has_prefix(str, "0x")) {
+			WacomStylusId id;
+			if (parse_stylus_id(str, &id)) {
+				WacomStylus *stylus = g_hash_table_lookup(db->stylus_ht, &id);
+				if (stylus)
+					g_array_append_val (array, stylus);
+				else
+					g_warning ("Invalid stylus id for '%s', ignoring stylus", str);
+			} else {
+				g_warning ("Invalid stylus id format for '%s', ignoring stylus", str);
 			}
-		} else if (g_str_has_prefix(id, "@")) {
-			const char *group = &id[1];
+		} else if (g_str_has_prefix(str, "@")) {
+			const char *group = &str[1];
 			GHashTableIter iter;
 			gpointer key, value;
 
@@ -721,17 +962,27 @@ libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
 			while (g_hash_table_iter_next (&iter, &key, &value)) {
 				WacomStylus *stylus = value;
 				if (stylus->group && g_str_equal(group, stylus->group)) {
-					g_array_append_val (array, stylus->id);
+					g_array_append_val (array, stylus);
 				}
 			}
 		} else {
-			g_warning ("Invalid prefix for '%s'!", id);
+			g_warning ("Invalid prefix for '%s', ignoring stylus", str);
 		}
 	}
 	/* Using groups means we don't get the styli in ascending order.
 	   Sort it so the output is predictable */
 	g_array_sort(array, styli_id_sort);
 	device->styli = array;
+
+	/* The legacy PID-only stylus id list */
+	device->deprecated_styli_ids = g_array_new(FALSE, FALSE, sizeof(int));
+	for (guint i = 0; i < device->styli->len; i++) {
+		WacomStylus *stylus = g_array_index(device->styli, WacomStylus*, i);
+		/* This only ever worked for Wacom styli, so let's keep that behavior */
+		if (stylus->id.vid == 0 || stylus->id.vid == WACOM_VENDOR_ID) {
+			g_array_append_val(device->deprecated_styli_ids, stylus->id.tool_id);
+		}
+	}
 }
 
 static void
@@ -871,7 +1122,7 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 				}
 			}
 			if (!found)
-				g_warning ("Unrecognized integration flag '%s'", string_list[i]);
+				g_warning ("Unrecognized integration flag '%s', ignoring flag", string_list[i]);
 		}
 		g_strfreev (string_list);
 	}
@@ -889,9 +1140,15 @@ libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
 
 	string_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "Styli", NULL, NULL);
 	if (!string_list) {
-		string_list = g_new0(char*, 3);
-		string_list[0] = g_strdup_printf("0x%x", WACOM_ERASER_FALLBACK_ID);
-		string_list[1] = g_strdup_printf("0x%x", WACOM_STYLUS_FALLBACK_ID);
+		GError *error = NULL;
+		if (g_key_file_get_boolean(keyfile, FEATURES_GROUP, "Stylus", &error) ||
+		    g_error_matches (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_KEY_NOT_FOUND)) {
+			string_list = g_new0(char*, 3);
+			string_list[0] = g_strdup_printf("0x0:0x%x", WACOM_ERASER_FALLBACK_ID);
+			string_list[1] = g_strdup_printf("0x0:0x%x", WACOM_STYLUS_FALLBACK_ID);
+		}
+		if (error)
+			g_error_free(error);
 	}
 	libwacom_parse_styli_list(db, device, string_list);
 	g_strfreev (string_list);
@@ -915,8 +1172,10 @@ out:
 		g_key_file_free(keyfile);
 	if (error)
 		g_error_free(error);
-	if (!success)
+	if (!success) {
+		g_warning("Ignoring invalid .tablet file %s", filename);
 		device = libwacom_unref(device);
+	}
 
 	return device;
 }
@@ -949,25 +1208,16 @@ is_stylus_file(const struct dirent *entry)
 }
 
 static bool
-load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
+load_tablet_files(WacomDeviceDatabase *db, GHashTable *parsed_filenames, const char *datadir)
 {
 	DIR *dir;
 	struct dirent *file;
 	bool success = false;
-	GHashTable *keyset = NULL;
 
 	dir = opendir(datadir);
 	if (!dir)
 		return errno == ENOENT; /* non-existing directory is ok */
 
-	/* A set of all matches for duplicate detection. We allow duplicates
-	 * across data directories, but we don't allow for duplicates
-	 * within the same data directory.
-	 */
-	keyset = g_hash_table_new_full (g_str_hash, g_str_equal, free, NULL);
-	if (!keyset)
-		goto out;
-
 	while ((file = readdir(dir))) {
 		WacomDevice *d;
 		guint idx = 0;
@@ -975,6 +1225,11 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 		if (!is_tablet_file(file))
 			continue;
 
+		if (g_hash_table_lookup(parsed_filenames, file->d_name))
+			continue;
+
+		g_hash_table_add(parsed_filenames, g_strdup(file->d_name));
+
 		d = libwacom_parse_tablet_keyfile(db, datadir, file->d_name);
 		if (!d)
 			continue;
@@ -991,28 +1246,13 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 			const char *matchstr;
 
 			matchstr = libwacom_match_get_match_string(match);
-			/* no duplicate matches allowed within the same
-			 * directory */
-			if (g_hash_table_contains(keyset, matchstr)) {
+			/* no duplicate matches allowed */
+			if (g_hash_table_contains(db->device_ht, matchstr)) {
 				g_critical("Duplicate match of '%s' on device '%s'.",
 					   matchstr, libwacom_get_name(d));
 				goto out;
 			}
-			g_hash_table_add(keyset, g_strdup(matchstr));
-
-			/* We already have an entry for this match in the database,
-			 * that takes precedence. Remove the current match
-			 * from the new tablet - that's fine because we
-			 * haven't exposed the tablet yet.
-			 */
-			if (g_hash_table_lookup(db->device_ht, matchstr)) {
-				libwacom_remove_match(d, match);
-				continue;
-			}
-
-			g_hash_table_insert(db->device_ht,
-					    g_strdup (matchstr),
-					    d);
+			g_hash_table_insert(db->device_ht, g_strdup (matchstr), d);
 			libwacom_ref(d);
 			idx++;
 		}
@@ -1022,8 +1262,6 @@ load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
 	success = true;
 
 out:
-	if (keyset)
-		g_hash_table_destroy(keyset);
 	closedir(dir);
 	return success;
 }
@@ -1035,7 +1273,7 @@ stylus_destroy(void *data)
 }
 
 static bool
-load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
+load_stylus_files(WacomDeviceDatabase *db, const char *datadir, AliasStatus handle_aliases)
 {
 	DIR *dir;
 	struct dirent *file;
@@ -1051,7 +1289,7 @@ load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
 			continue;
 
 		path = g_build_filename (datadir, file->d_name, NULL);
-		libwacom_parse_stylus_keyfile(db, path);
+		libwacom_parse_stylus_keyfile(db, path, handle_aliases);
 		g_free(path);
 	}
 
@@ -1060,37 +1298,63 @@ load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
 	return true;
 }
 
+static guint
+stylus_hash(WacomStylusId *id)
+{
+	guint64 full_id = (guint64)id->vid << 32 | id->tool_id;
+	return g_int64_hash(&full_id);
+}
+
+static gboolean
+stylus_compare(WacomStylusId *a, WacomStylusId *b)
+{
+	return wacom_stylus_id_sort(a, b) == 0;
+}
+
 static WacomDeviceDatabase *
-database_new_for_paths (size_t npaths, const char **datadirs)
+database_new_for_paths (char * const *datadirs)
 {
 	WacomDeviceDatabase *db;
-	size_t n;
-	const char **datadir;
+	char * const *datadir;
+	GHashTable *parsed_filenames;
+
+	parsed_filenames = g_hash_table_new_full (g_str_hash, g_str_equal, free, NULL);
+	if (!parsed_filenames)
+		return NULL;
 
 	db = g_new0 (WacomDeviceDatabase, 1);
 	db->device_ht = g_hash_table_new_full (g_str_hash,
 					       g_str_equal,
 					       g_free,
 					       (GDestroyNotify) libwacom_destroy);
-	db->stylus_ht = g_hash_table_new_full (g_direct_hash,
-					       g_direct_equal,
-					       NULL,
+	db->stylus_ht = g_hash_table_new_full ((GHashFunc)stylus_hash,
+					       (GEqualFunc)stylus_compare,
+					       (GDestroyNotify) g_free,
 					       (GDestroyNotify) stylus_destroy);
 
-	for (datadir = datadirs, n = npaths; n--; datadir++) {
-		if (!load_stylus_files(db, *datadir))
+	for (datadir = datadirs; *datadir; datadir++) {
+		if (!load_stylus_files(db, *datadir, IGNORE_ALIASES))
 			goto error;
 	}
 
-	for (datadir = datadirs, n = npaths; n--; datadir++) {
-		if (!load_tablet_files(db, *datadir))
+	for (datadir = datadirs; *datadir; datadir++) {
+		if (!load_stylus_files(db, *datadir, ONLY_ALIASES))
 			goto error;
 	}
 
+	for (datadir = datadirs; *datadir;  datadir++) {
+		if (!load_tablet_files(db, parsed_filenames, *datadir))
+			goto error;
+	}
+
+	g_hash_table_unref(parsed_filenames);
+
 	/* If we couldn't load _anything_ then something's wrong */
 	if (g_hash_table_size (db->stylus_ht) == 0 ||
-	    g_hash_table_size (db->device_ht) == 0)
+	    g_hash_table_size (db->device_ht) == 0) {
+		g_warning("Zero tablet or stylus files found in datadirs");
 		goto error;
+	}
 
 	libwacom_setup_paired_attributes(db);
 
@@ -1104,18 +1368,42 @@ error:
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new_for_path (const char *datadir)
 {
-	return database_new_for_paths(1, &datadir);
+	WacomDeviceDatabase *db;
+	char **paths;
+
+	paths = g_strsplit(datadir, ":", 0);
+	db = database_new_for_paths(paths);
+
+	g_strfreev(paths);
+
+	return db;
 }
 
 LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new (void)
 {
-	const char *datadir[] = {
+	WacomDeviceDatabase *db;
+	char *xdgdir = NULL;
+	char *xdg_config_home = g_strdup(g_getenv("XDG_CONFIG_HOME"));
+
+	if (!xdg_config_home)
+		xdg_config_home = g_strdup_printf("%s/.config/", g_get_home_dir());
+
+	xdgdir = g_strdup_printf("%s/libwacom", xdg_config_home);
+
+	char *datadir[] = {
+		xdgdir,
 		ETCDIR,
 		DATADIR,
+		NULL,
 	};
 
-	return database_new_for_paths (2, datadir);
+	db = database_new_for_paths(datadir);
+
+	free(xdgdir);
+	free(xdg_config_home);
+
+	return db;
 }
 
 LIBWACOM_EXPORT void
