@@ -137,6 +137,10 @@ get_bus_vid_pid (GUdevDevice  *device,
 	*product_id = (int)strtol (splitted_product[2], NULL, 16);
 
 	switch (bus_id) {
+	case 0:
+		*bus = WBUSTYPE_UNKNOWN;
+		retval = TRUE;
+		break;
 	case 3:
 		*bus = WBUSTYPE_USB;
 		retval = TRUE;
@@ -419,7 +423,9 @@ libwacom_copy(const WacomDevice *device)
 	d->ring_num_modes = device->ring_num_modes;
 	d->ring2_num_modes = device->ring2_num_modes;
 	d->styli = g_array_copy(device->styli);
+	d->deprecated_styli_ids = g_array_copy(device->deprecated_styli_ids);
 	d->status_leds = g_array_copy(device->status_leds);
+
 	d->buttons = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 					   NULL, g_free);
 	g_hash_table_iter_init(&iter, device->buttons);
@@ -544,11 +550,23 @@ libwacom_compare(const WacomDevice *a, const WacomDevice *b, WacomCompareFlags f
 	if (g_hash_table_size(a->buttons) != g_hash_table_size(b->buttons))
 		return 1;
 
+	/* We don't need to check deprecated_stylus_ids because if they differ
+	 * when the real id doesn't that's a bug */
+
 	if (a->styli->len != b->styli->len)
 		return 1;
 
-	if (memcmp(a->styli->data, b->styli->data, sizeof(int) * a->styli->len) != 0)
-		return 1;
+	/* This needs to be a deep comparison - our styli array contains
+	 * WacomStylus* pointers but we want libwacom_compare() to return
+	 * true if the stylus data matches (test-dbverify compares styli
+	 * from two different WacomDeviceDatabase).
+	 */
+	for (guint i = 0; i < a->styli->len; i++) {
+		WacomStylus *as = g_array_index(a->styli, WacomStylus*, i);
+		WacomStylus *bs = g_array_index(b->styli, WacomStylus*, i);
+		if (as->id.tool_id != bs->id.tool_id)
+			return 1;
+	}
 
 	if (a->status_leds->len != b->status_leds->len)
 		return 1;
@@ -849,7 +867,7 @@ libwacom_new_from_name(const WacomDeviceDatabase *db, const char *name, WacomErr
 static void print_styli_for_device (int fd, const WacomDevice *device)
 {
 	int nstyli;
-	const int *styli;
+	const WacomStylus **styli;
 	int i;
 	unsigned idx = 0;
 	char buf[1024] = {0};
@@ -857,14 +875,18 @@ static void print_styli_for_device (int fd, const WacomDevice *device)
 	if (!libwacom_has_stylus(device))
 		return;
 
-	styli = libwacom_get_supported_styli(device, &nstyli);
-
+	styli = libwacom_get_styli(device, &nstyli);
 	for (i = 0; i < nstyli; i++) {
+		const WacomStylus *stylus = styli[i];
 		/* 20 digits for a stylus are enough, right */
 		assert(idx < sizeof(buf) - 20);
 
-		idx += snprintf(buf + idx, 20, "%#x;", styli[i]);
+		if (stylus->id.vid != WACOM_VENDOR_ID)
+			idx += snprintf(buf + idx, 20, "0x%04x:%#x;", stylus->id.vid, stylus->id.tool_id);
+		else
+			idx += snprintf(buf + idx, 20, "%#x;", stylus->id.tool_id);
 	}
+	g_free(styli);
 
 	dprintf(fd, "Styli=%s\n", buf);
 }
@@ -887,8 +909,8 @@ static void print_supported_leds (int fd, const WacomDevice *device)
 	char *leds_name[] = {
 		"Ring;",
 		"Ring2;",
-		"Touchstrip;",
-		"Touchstrip2;"
+		"Strip;",
+		"Strip2;"
 	};
 	int num_leds;
 	const WacomStatusLEDs *status_leds;
@@ -959,8 +981,8 @@ static void print_buttons_for_device (int fd, const WacomDevice *device)
 	print_button_flag_if(fd, device, "Right", WACOM_BUTTON_POSITION_RIGHT);
 	print_button_flag_if(fd, device, "Top", WACOM_BUTTON_POSITION_TOP);
 	print_button_flag_if(fd, device, "Bottom", WACOM_BUTTON_POSITION_BOTTOM);
-	print_button_flag_if(fd, device, "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
-	print_button_flag_if(fd, device, "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH);
+	print_button_flag_if(fd, device, "Strip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
+	print_button_flag_if(fd, device, "Strip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH);
 	print_button_flag_if(fd, device, "Dial", WACOM_BUTTON_DIAL_MODESWITCH);
 	print_button_flag_if(fd, device, "OLEDs", WACOM_BUTTON_OLED);
 	print_button_flag_if(fd, device, "Ring", WACOM_BUTTON_RING_MODESWITCH);
@@ -991,6 +1013,8 @@ static void print_integrated_flags_for_device (int fd, const WacomDevice *device
 		dprintf(fd, "Display;");
 	if (device->integration_flags & WACOM_DEVICE_INTEGRATED_SYSTEM)
 		dprintf(fd, "System;");
+	if (device->integration_flags & WACOM_DEVICE_INTEGRATED_REMOTE)
+		dprintf(fd, "Remote;");
 	dprintf(fd, "\n");
 }
 
@@ -1068,7 +1092,7 @@ libwacom_print_device_description(int fd, const WacomDevice *device)
 	dprintf(fd, "[Features]\n");
 	dprintf(fd, "Reversible=%s\n", libwacom_is_reversible(device)	? "true" : "false");
 	dprintf(fd, "Stylus=%s\n",	 libwacom_has_stylus(device)	? "true" : "false");
-	dprintf(fd, "NumRings=%s\n",	 libwacom_get_num_rings(device)	? "true" : "false");
+	dprintf(fd, "NumRings=%d\n",	 libwacom_get_num_rings(device));
 	dprintf(fd, "Touch=%s\n",	 libwacom_has_touch(device)	? "true" : "false");
 	dprintf(fd, "TouchSwitch=%s\n",	libwacom_has_touchswitch(device)? "true" : "false");
 	print_supported_leds(fd, device);
@@ -1112,6 +1136,7 @@ libwacom_unref(WacomDevice *device)
 	g_clear_pointer (&device->matches, g_array_unref);
 	libwacom_match_unref(device->match);
 	g_clear_pointer (&device->styli, g_array_unref);
+	g_clear_pointer (&device->deprecated_styli_ids, g_array_unref);
 	g_clear_pointer (&device->status_leds, g_array_unref);
 	g_clear_pointer (&device->buttons, g_hash_table_destroy);
 	g_free (device);
@@ -1343,7 +1368,20 @@ libwacom_get_height(const WacomDevice *device)
 LIBWACOM_EXPORT WacomClass
 libwacom_get_class(const WacomDevice *device)
 {
-	return device->cls;
+	if (device->cls != WCLASS_UNKNOWN) {
+		return device->cls;
+	}
+
+	switch (device->integration_flags) {
+	case WACOM_DEVICE_INTEGRATED_DISPLAY:
+		return WCLASS_CINTIQ;
+	case WACOM_DEVICE_INTEGRATED_DISPLAY|WACOM_DEVICE_INTEGRATED_SYSTEM:
+		return WCLASS_CINTIQ;
+	case WACOM_DEVICE_INTEGRATED_REMOTE:
+		return WCLASS_REMOTE;
+	}
+
+	return WCLASS_BAMBOO;
 }
 
 LIBWACOM_EXPORT int
@@ -1373,8 +1411,23 @@ libwacom_get_num_keys(const WacomDevice *device)
 LIBWACOM_EXPORT const int *
 libwacom_get_supported_styli(const WacomDevice *device, int *num_styli)
 {
-	*num_styli = device->styli->len;
-	return (const int *)device->styli->data;
+	*num_styli = device->deprecated_styli_ids->len;
+	return (const int *)device->deprecated_styli_ids->data;
+}
+
+LIBWACOM_EXPORT const WacomStylus **
+libwacom_get_styli(const WacomDevice *device, int *num_styli)
+{
+	int count = device->styli->len;
+	const WacomStylus **styli = g_new0(const WacomStylus*, count + 1);
+
+	if (count > 0)
+		memcpy(styli, device->styli->data, count * sizeof(WacomStylus*));
+
+	if (num_styli)
+		*num_styli = count;
+
+	return styli;
 }
 
 LIBWACOM_EXPORT int
@@ -1451,7 +1504,9 @@ static const struct {
 	{ WACOM_BUTTON_RING_MODESWITCH,		WACOM_STATUS_LED_RING },
 	{ WACOM_BUTTON_RING2_MODESWITCH,	WACOM_STATUS_LED_RING2 },
 	{ WACOM_BUTTON_TOUCHSTRIP_MODESWITCH,	WACOM_STATUS_LED_TOUCHSTRIP },
-	{ WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH,	WACOM_STATUS_LED_TOUCHSTRIP2 }
+	{ WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH,	WACOM_STATUS_LED_TOUCHSTRIP2 },
+	{ WACOM_BUTTON_DIAL_MODESWITCH,		WACOM_STATUS_LED_DIAL },
+	{ WACOM_BUTTON_DIAL2_MODESWITCH,	WACOM_STATUS_LED_DIAL2 },
 };
 
 LIBWACOM_EXPORT int
@@ -1477,7 +1532,7 @@ libwacom_get_button_led_group (const WacomDevice *device, char button)
 		}
 	}
 
-	return WACOM_STATUS_LED_UNAVAILABLE;
+	return -1;
 }
 
 LIBWACOM_EXPORT int
@@ -1534,16 +1589,54 @@ libwacom_get_button_evdev_code(const WacomDevice *device, char button)
 	return b ? b->code : 0;
 }
 
-LIBWACOM_EXPORT const WacomStylus *
-libwacom_stylus_get_for_id (const WacomDeviceDatabase *db, int id)
+LIBWACOM_EXPORT WacomModeSwitch
+libwacom_get_button_modeswitch_mode(const WacomDevice *device, char button)
 {
-	return g_hash_table_lookup (db->stylus_ht, GINT_TO_POINTER(id));
+	WacomButton *b = g_hash_table_lookup(device->buttons,
+					     GINT_TO_POINTER(button));
+
+	if (!b || (b->flags & WACOM_BUTTON_MODESWITCH) == 0)
+		return WACOM_MODE_SWITCH_NEXT;
+
+	return b->mode;
+}
+
+static const WacomStylus *
+libwacom_stylus_get_for_stylus_id (const WacomDeviceDatabase *db,
+				   const WacomStylusId *id)
+{
+	return g_hash_table_lookup (db->stylus_ht, id);
+}
+
+LIBWACOM_EXPORT const WacomStylus *
+libwacom_stylus_get_for_id (const WacomDeviceDatabase *db, int tool_id)
+{
+	WacomStylusId id = {
+		.vid = WACOM_VENDOR_ID,
+		.tool_id = tool_id,
+	};
+
+	switch (tool_id) {
+	case GENERIC_PEN_WITH_ERASER:
+	case GENERIC_ERASER:
+	case GENERIC_PEN_NO_ERASER:
+		id.vid = 0;
+		break;
+
+	}
+	return libwacom_stylus_get_for_stylus_id (db, &id);
 }
 
 LIBWACOM_EXPORT int
 libwacom_stylus_get_id (const WacomStylus *stylus)
 {
-	return stylus->id;
+	return stylus->id.tool_id;
+}
+
+LIBWACOM_EXPORT int
+libwacom_stylus_get_vendor_id (const WacomStylus *stylus)
+{
+	return stylus->id.vid;
 }
 
 LIBWACOM_EXPORT const char *
@@ -1556,15 +1649,29 @@ LIBWACOM_EXPORT const int *
 libwacom_stylus_get_paired_ids(const WacomStylus *stylus, int *num_paired_ids)
 {
 	if (num_paired_ids)
-		*num_paired_ids = stylus->paired_ids->len;
-	return (const int*)stylus->paired_ids->data;
+		*num_paired_ids = stylus->deprecated_paired_ids->len;
+	return (const int*)stylus->deprecated_paired_ids->data;
+}
+
+LIBWACOM_EXPORT const WacomStylus **
+libwacom_stylus_get_paired_styli(const WacomStylus *stylus, int *num_paired)
+{
+	int count = stylus->paired_styli->len;
+	const WacomStylus **styli = g_new0(const WacomStylus*, count + 1);
+
+	if (num_paired)
+		*num_paired = count;
+
+	if (count > 0)
+		memcpy(styli, stylus->paired_styli->data, count * sizeof(WacomStylus*));
+	return styli;
 }
 
 LIBWACOM_EXPORT int
 libwacom_stylus_get_num_buttons (const WacomStylus *stylus)
 {
 	if (stylus->num_buttons == -1) {
-		g_warning ("Stylus '0x%x' has no number of buttons defined, falling back to 2", stylus->id);
+		g_warning ("Stylus '0x%x' has no number of buttons defined, falling back to 2", stylus->id.tool_id);
 		return 2;
 	}
 	return stylus->num_buttons;
@@ -1604,7 +1711,7 @@ LIBWACOM_EXPORT WacomStylusType
 libwacom_stylus_get_type (const WacomStylus *stylus)
 {
 	if (stylus->type == WSTYLUS_UNKNOWN) {
-		g_warning ("Stylus '0x%x' has no type defined, falling back to 'General'", stylus->id);
+		g_warning ("Stylus '0x%x' has no type defined, falling back to 'General'", stylus->id.tool_id);
 		return WSTYLUS_GENERAL;
 	}
 	return stylus->type;
@@ -1621,17 +1728,25 @@ libwacom_print_stylus_description (int fd, const WacomStylus *stylus)
 {
 	const char *type;
 	WacomAxisTypeFlags axes;
-	const int *paired_ids;
+	const WacomStylus **paired;
 	int count;
 	int i;
 
-	dprintf(fd, "[%#x]\n",		libwacom_stylus_get_id(stylus));
+	if (libwacom_stylus_get_vendor_id(stylus) != WACOM_VENDOR_ID)
+		dprintf(fd, "[0x%x:%#x]\n", libwacom_stylus_get_vendor_id(stylus), libwacom_stylus_get_id(stylus));
+	else
+		dprintf(fd, "[%#x]\n", libwacom_stylus_get_id(stylus));
+
 	dprintf(fd, "Name=%s\n",	libwacom_stylus_get_name(stylus));
 	dprintf(fd, "PairedIds=");
-	paired_ids = libwacom_stylus_get_paired_ids(stylus, &count);
+	paired = libwacom_stylus_get_paired_styli(stylus, &count);
 	for (i = 0; i < count; i++) {
-		dprintf(fd, "%#x;", paired_ids[i]);
+		if (paired[i]->id.vid != 0x56a)
+			dprintf(fd, "%#04x:%#x;", paired[i]->id.vid, paired[i]->id.tool_id);
+		else
+			dprintf(fd, "%#x;", paired[i]->id.tool_id);
 	}
+	g_free(paired);
 	dprintf(fd, "\n");
 	switch (libwacom_stylus_get_eraser_type(stylus)) {
 		case WACOM_ERASER_UNKNOWN: type = "Unknown";       break;
@@ -1690,7 +1805,9 @@ libwacom_stylus_unref(WacomStylus *stylus)
 
 	g_free (stylus->name);
 	g_free (stylus->group);
-	g_clear_pointer (&stylus->paired_ids, g_array_unref);
+	g_clear_pointer (&stylus->deprecated_paired_ids, g_array_unref);
+	g_clear_pointer (&stylus->paired_stylus_ids, g_array_unref);
+	g_clear_pointer (&stylus->paired_styli, g_array_unref);
 	g_free (stylus);
 
 	return NULL;
